/*  flip SDL clock
******************************************************************
  fc3.c
  Nokia N810
  
  gcc -g -Wall `sdl-config --cflags` `sdl-config --libs` -L /usr/include/SDL -lSDL_image fc3.c -o fc
  Ciro Ippolito 2009
  ******************************************************************
*/

/*2DO
one png for the entire number sets
...
...
********************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/select.h>
#include <unistd.h>
#include <fcntl.h>

#include "SDL_image.h"

#if defined(_MSC_VER)
#include "SDL.h"
#else
#include "SDL/SDL.h"
#endif

static int display_active = 1;
const int LOOP_DELAY = 50000; //.05sec

SDL_Surface *screen;
SDL_Surface *secs;
SDL_Surface *fondo;
SDL_Surface *zero;
SDL_Surface *uno;
SDL_Surface *due;
SDL_Surface *tre;
SDL_Surface *quattro;
SDL_Surface *cinque;
SDL_Surface *sei;
SDL_Surface *sette;
SDL_Surface *otto;
SDL_Surface *nove;
SDL_Surface *numbers[10];
SDL_Surface *elements[3];

#define CLOCKID CLOCK_MONOTONIC
#define SIG SIGALRM
#define FC_TIMER  		201
#define FC_TIMEOUT  	100 //1 secs 


#include "SFont.h"
extern int SFont_DefaultCharWidth;

SFont_Font* Font=NULL;

int fontheight = 30;

SDL_Surface* fade;


#ifdef ZIPIT_Z2
#define SCREENWIDTH 320
#define SCREENHEIGHT 240
#define SCREENBPP 32
#else
#define SCREENWIDTH 800
#define SCREENHEIGHT 480
#define SCREENBPP 16
#endif

const int CLOCK_TOP = (SCREENHEIGHT-152)/2;

void UpdateClock();

static void onTimer(int sig, siginfo_t *si, void *uc)
{
	switch(si->si_int){
		case FC_TIMER:{
			SDL_Event event;

			event.type = SDL_USEREVENT;
			event.user.code = 0;
			event.user.data1 = &UpdateClock;
			event.user.data2 = 0;
			
			SDL_PushEvent(&event);
			
			break;
		}
		default:
			break;
	}
}

		
timer_t create_timer(int timerName, unsigned int freq_msecs)
{
    struct itimerspec 	its;
						its.it_value.tv_sec = freq_msecs / 100;
						its.it_value.tv_nsec = 0;
						its.it_interval.tv_sec = 0;
						its.it_interval.tv_nsec = 0;
						
	struct sigevent 	sev;
						sev.sigev_notify = SIGEV_SIGNAL;
						sev.sigev_signo = SIG;
						sev.sigev_value.sival_int = timerName;

	struct sigaction 	sa;
						sa.sa_flags = SA_SIGINFO;
						sa.sa_sigaction = onTimer;

	timer_t timerid=0;

	/* Establish handler for timer signal */
	sigemptyset(&sa.sa_mask);
	if (sigaction(SIG, &sa, NULL) == -1){
		fprintf(stderr, "create_timer failed sigaction ");
		exit(1);
	}
	/* Create the timer */
	if (timer_create(CLOCKID, &sev, &timerid) == -1){
		fprintf(stderr, "create_timer failed timer_create ");
		exit(1);
	}

   /* Start the timer */
   if (timer_settime(timerid, 0, &its, NULL) == -1){
		fprintf(stderr, "create_timer failed timer_settime ");
		exit(1);
	}
   
   return timerid;
}

static int set_timer(timer_t timerid, unsigned int freq_msecs)
{
    struct itimerspec 	its;
						its.it_value.tv_sec = freq_msecs / 100;
						its.it_value.tv_nsec = 0;
						its.it_interval.tv_sec = 0;
						its.it_interval.tv_nsec = 0;
	
   /* Start the timer */
   if (timer_settime(timerid, 0, &its, NULL) == -1){
		fprintf(stderr, "set_timer failed timer_settime ");
		exit(1);
	}
   
   return 1;
}

int is_ready(int fd) {
    fd_set fdset;
    struct timeval timeout;
    int ret;
    FD_ZERO(&fdset);
    FD_SET(fd, &fdset);
    timeout.tv_sec = 0;
    timeout.tv_usec = 1;
    return select(fd+1, &fdset, NULL, NULL, &timeout) == 1 ? 1 : 0;
}

SDL_Surface *loadImage(char *filename)
{
	char buf[50];
	snprintf(buf, sizeof buf, "/usr/share/fc/%s", filename);
    // Load Picture
    SDL_Surface *tmp = IMG_Load(buf);
    if (tmp == (SDL_Surface *) (NULL)) 
    {
        fprintf(stderr, "\n Couldn't load image file %s: %s\n\n", filename, SDL_GetError());
        SDL_Quit();
    }
	
    return tmp;
}

enum{
	FONDO=0,
	BARRA,
	SECS
};

void initFont()
{
    // WARNING!!  Gotta fix sfont for utf8 bits or this crashes on gmenu2x font.
    //            Maybe just stop at first utf8 (190?)

    //Font = SFont_InitFont(IMG_Load("font.png"));
    SDL_Surface* font = loadImage("font.png");
    Font = SFont_InitFont(font);
    if(!Font) {
		fprintf(stderr, "An error occured while loading the font.\n");
		exit(1);
    }
    fontheight = SFont_TextHeight(Font);

    // Create a surface with alpha to use for fade effect.
    SDL_Surface* fadeSurface = SDL_CreateRGBSurface( SDL_SWSURFACE | SDL_SRCALPHA, 320, 240, 32, 0xff000000,0x00ff0000,0x0000ff00,0x000000ff);

    // Convert it to the format of the screen
    fade = SDL_DisplayFormat( fadeSurface );
    SDL_SetAlpha(fade, SDL_SRCALPHA | SDL_RLEACCEL, 128);

    // Free the created surface
    SDL_FreeSurface( fadeSurface );	
	
}
void initBMP()
{
    elements[FONDO] = loadImage("fondo.png");
	elements[BARRA] = loadImage("barra.png");
    elements[SECS] = loadImage("secs.png");
    numbers[0]  = loadImage("0.png");
    numbers[1]  = loadImage("1.png");
    numbers[2]  = loadImage("2.png");
    numbers[3]  = loadImage("3.png");
    numbers[4]  = loadImage("4.png");
    numbers[5]  = loadImage("5.png");
    numbers[6]  = loadImage("6.png");
    numbers[7]  = loadImage("7.png");
    numbers[8]  = loadImage("8.png");
    numbers[9]  = loadImage("9.png");
}

void BMPShow(SDL_Surface *screen, SDL_Surface *pic,int x, int y)
{
  SDL_Rect dest;
  dest.x = x;
  dest.y = y;
  dest.w = pic->w;
  dest.h = pic->h;
  SDL_BlitSurface(pic, NULL, screen, &dest);
  SDL_Flip(screen);
    // Update the changed portion of the screen 
//    SDL_UpdateRects(screen, 1, &dest);
  
  return;
}

void set_clock(struct tm* time){
	
	int min_p1 = time->tm_min/10;
	int min_p2 = time->tm_min - min_p1*10;
	int hr = time->tm_hour>12?time->tm_hour-12:time->tm_hour;
	if(!hr) hr = 12;
	int hr_p1 = hr/10;
	int hr_p2 = hr - hr_p1*10;
	const int NUMBER_TOP=CLOCK_TOP+elements[BARRA]->h;
	const int NUMBER_WIDTH = numbers[0]->w;
#ifdef ZIPIT_Z2				
	BMPShow (screen,numbers[hr_p1] , 0, NUMBER_TOP);
	BMPShow (screen,numbers[hr_p2] , NUMBER_WIDTH, NUMBER_TOP);
	BMPShow (screen,numbers[min_p1] , SCREENWIDTH-2*NUMBER_WIDTH, NUMBER_TOP);
	BMPShow (screen,numbers[min_p2] , SCREENWIDTH-NUMBER_WIDTH, NUMBER_TOP);
#else
	BMPShow (screen,numbers[hr_p1] , 2,20);
	BMPShow (screen,numbers[hr_p2] , 193,20);
	BMPShow (screen,numbers[min_p1] , 415,20);
	BMPShow (screen,numbers[min_p2] , 605,20);
#endif	
}

char *bannertext[4] = {NULL,NULL,"%b %d, %Y  -  %l:%M %p",NULL};
char bannertype[4] = {'t','t','D','t'};
int bannerfade[4] = {1,1,1,1};
int bannerfile[4] = {-1,-1,-1,-1};
int bannerfpos[4] = {0,0,0,0};
char *bannerfstr[4] = {NULL,NULL,NULL,NULL};

void UpdateClock ()
{
	if (display_active){    
		time_t now = time(NULL);
		struct tm* tim = localtime(&now);
		int secs = tim->tm_sec;  
		if(secs==0)	set_clock(tim);

		BMPShow (screen,elements[BARRA] ,0   , CLOCK_TOP);            	//Redraw the top minute bar
#ifdef ZIPIT_Z2
		BMPShow (screen,elements[SECS] , (secs*52/10)+4 , CLOCK_TOP);  	//Seconds arrow
#else
		BMPShow (screen,elements[SECS] , secs*13+10 , CLOCK_TOP);     	//Seconds arrow
#endif

		if (Font)
		{
			size_t i,n;
			SDL_Rect dest;
			dest.x = 0;
			dest.y = 150 + 2 + 22-fontheight;
			dest.w = 320; //SFont_TextWidth(Font, s);
			dest.h = 22;
			for (n=0; n<4; n++)
			{

				if (bannertext[n] && strlen(bannertext[n]))
				{
					//SFont_Write(screen, Font, 0, dest.y, bannertext); 
					if (bannertype[n] == 'f')
					{
						int fd = bannerfile[n];
						if (fd < 0)
							fd = open(bannertext[n], O_RDONLY);
						bannerfile[n] = fd;
						if (fd >= 0)
						{
							char buffer[128];
							char *p;
							int j = bannerfpos[n];
							buffer[0] = 0;
							if (is_ready(fd)) 
							{
								j += read(fd, buffer, sizeof(buffer)-2-j);
								buffer[j] = 0;
								if (j == sizeof(buffer)-2)
									strcat(buffer, "\n");
								if (p = strpbrk(buffer, "\r\n")){
									*p = 0;
									close(fd);
									bannerfile[n] =-1;
									bannerfpos[n] = 0;
									if (bannerfstr[n]) free(bannerfstr[n]);
										bannerfstr[n] = strdup(buffer);
								}
							}
						}
						SDL_FillRect(screen, &dest, 0 );
						if (bannerfstr[n])
							SFont_WriteCenter(screen, Font, dest.y, bannerfstr[n]);
					}
					else if (bannertype[n] == 'D')
					{
						char s[30];
						strftime(s,30,bannertext[n],tim);
						//printf("Clock text test %s\n",s);
						SDL_FillRect(screen, &dest, 0 );
						SFont_WriteCenter(screen,Font,dest.y,s);
					}										
					else
						SFont_WriteCenter(screen, Font, dest.y, bannertext[n]);
					for (i=0; i<bannerfade[n]; i++)
						SDL_BlitSurface(fade, &dest, screen, &dest); // Fade it.
				}
				dest.y += dest.h;
			}
		//	SDL_Flip(screen);
		}
	}
}
/* ----------------------------------------------------------------------------
MAIN ()
 ----------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
    if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) 
    {
        fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
        exit(1);
    }
	
	timer_t fc_timerid = create_timer(FC_TIMER, FC_TIMEOUT);
	set_timer(fc_timerid, FC_TIMEOUT);

    // Register SDL_Quit to be called at exit; makes sure things are
    atexit(SDL_Quit); 								// cleaned up when we quit.

    screen = SDL_SetVideoMode(SCREENWIDTH, SCREENHEIGHT, SCREENBPP, SDL_HWSURFACE|SDL_DOUBLEBUF);                   // Window PC

    //screen = SDL_SetVideoMode(800, 480, 16, SDL_SWSURFACE|SDL_FULLSCREEN);  // Full screen Tablet
    screen = SDL_SetVideoMode(SCREENWIDTH, SCREENHEIGHT, SCREENBPP, SDL_SWSURFACE);                   // Window PC

    if ( screen == NULL )     							                      // If fail, return error
    {
        fprintf(stderr, "Unable to set 640x480 video: %s\n", SDL_GetError());
        exit(1);
    }

	SDL_ShowCursor(SDL_DISABLE);

	initFont();
    initBMP();
	
    BMPShow (screen, elements[FONDO] ,0   , CLOCK_TOP+elements[BARRA]->h);        //Fondo
    BMPShow (screen, elements[BARRA] ,0   , CLOCK_TOP);        //barra minuti

	time_t now = time(NULL);
	struct tm* tim = localtime(&now);
	set_clock(tim);

    typedef enum { FALSE=0, TRUE } boolean;
    boolean clockticking = TRUE;
	
    while (clockticking)					        // Main loop: loop forever
    {
        SDL_Event event;			                // Poll for events
		if(SDL_PollEvent(&event)) {
            switch (event.type) {
			case SDL_USEREVENT:{
				/* and now we can call the function we wanted to call in the timer but couldn't because of the multithreading problems */
				void (*p)(void) = event.user.data1;
				p();
				set_timer(fc_timerid, FC_TIMEOUT);
				break;
			}
			case SDL_ACTIVEEVENT:
				if (SDL_GetAppState() & SDL_APPACTIVE){
					display_active = 1;
					time_t now = time(NULL);
					struct tm* tim = localtime(&now);							
					set_clock(tim);
				}
				else
					display_active = 0;
				
				break;
            case SDL_KEYDOWN:
#ifndef ZIPIT_Z2				
				if (event.key.keysym.sym == SDLK_a)	{
					printf("premuto A");
				}
				else if (event.key.keysym.sym == SDLK_s){
					printf("premuto S");
				}
				else if (event.key.keysym.sym == SDLK_d){
					printf("premuto D");
				}
				else if (event.key.keysym.sym == SDLK_f){
					printf("premuto F");
				}
				else
#endif
				if (event.key.keysym.sym == SDLK_g)
					BMPShow (screen, elements[FONDO]  ,0    , CLOCK_TOP+elements[BARRA]->h);
				break;
			case SDL_KEYUP:
				if (!(event.key.keysym.sym == SDLK_ESCAPE || event.key.keysym.sym == SDLK_q))		// If Q is pressed, return (and zen quit)
					break;
				//else fall thru to default break
			case SDL_QUIT:
                clockticking=FALSE;
			default: 
				break;
            }
        }
		usleep(LOOP_DELAY);	
		
    }

    return 0;
}


